@@ -1,6 +1,6 @@
 import re
 from enum import Enum
-from typing import List, Callable, Tuple
+from typing import List, Callable, Tuple, Dict
 
 from . import Utilities, InstrumentSettings, InstrumentOptions, InstrumentErrors, Conversions as Conv
 from .ArgSingleSuppressed import ArgSingleSuppressed
@@ -11,6 +11,8 @@ from .StreamReader import StreamReader
 from .StreamWriter import StreamWriter
 from .Utilities import trim_str_response
 from .VisaSession import VisaSession, EventArgsChunk
+from .VisaSessionSim import VisaSessionSim
+from .RepeatedCapability import RepeatedCapability
 
 
 class Instrument(object):
@@ -21,7 +23,7 @@ class Instrument(object):
 		If simulate is true, it cannot be later switched to false anymore."""
 		self._simulating: bool = simulate
 		self._session = None
-		self._global_repcaps = {}
+		self._global_repcaps: Dict[str, RepeatedCapability] = {}
 		self._linker = InternalLinker()
 		# noinspection PyTypeChecker
 		self.on_write_handler: Callable = None
@@ -32,10 +34,11 @@ class Instrument(object):
 		# Fixed settings
 		self.selftest_timeout = 100000 if settings.selftest_timeout == 0 else settings.selftest_timeout
 		self.idn_model_full_name = settings.idn_model_full_name
+		self.instr_options_parse_mode = settings.instr_options_parse_mode
 
 		# Changeable settings
 		self.resource_name: str = resource_name
-		self.instr_options: InstrumentOptions = None
+		self._instr_options: InstrumentOptions = None  # Internal private property for the lazy property self.instr_options - see the getter for self.instr_options. Initialized by the first access
 		self.query_instr_status: bool = True
 		self.opc_query_after_write: bool = False
 		self.bin_float_numbers_format = settings.bin_float_numbers_format
@@ -49,6 +52,9 @@ class Instrument(object):
 		self.firmware_version: str = '1.00'
 
 		if self._simulating:
+			self._session = VisaSessionSim(resource_name, settings, direct_session)
+			self._instr_options = InstrumentOptions.Options('K0', InstrumentOptions.ParseMode.KeepOriginal)
+			self._session.write("*OPT K0")
 			return
 
 		self._session = VisaSession(resource_name, settings, direct_session)
@@ -59,8 +65,7 @@ class Instrument(object):
 		if self._session.is_rsnrp_session():
 			settings.instr_options_parse_mode = InstrumentOptions.ParseMode.Skip
 			self.stb_in_error_check = False
-
-		self._query_options_and_parse(settings.instr_options_parse_mode)
+		self.instr_options_parse_mode = settings.instr_options_parse_mode
 
 	def __str__(self):
 		if self._simulating:
@@ -68,6 +73,24 @@ class Instrument(object):
 		else:
 			return f"Instrument Model: '{self.model}', ResourceName: '{self.resource_name}'"
 
+	def set_simulating_cmds(self) -> None:
+		"""Updated cached values in the simulating VISA session to properly respond to *IDN? or *OPT?"""
+		if self._simulating:
+			self.write(f'*idn {self.idn_string}')
+			self.write(f'*opt {Conv.list_to_csv_str(self.instr_options.get_all())}')
+			self.write(f'*opc 1')
+			self.write(f'*stb 0')
+			self.write(f'*tst 0,"Passed"')
+			self.write(f'syst:err 0,"No Error"')
+			self.write(f'system:error 0,"No Error"')
+
+	def get_last_sent_cmd(self) -> str:
+		"""Returns the last commands sent to the instrument. Only works in simulation mode"""
+		if self._simulating:
+			# noinspection PyUnresolvedReferences
+			return self._session.get_last_sent_cmd()
+		raise Exception("get_last_sent_cmd() can only be used in simulation mode")
+
 	@property
 	def visa_manufacturer(self) -> str:
 		"""Returns the visa manufacturer of the current session."""
@@ -99,6 +122,13 @@ class Instrument(object):
 		self._idn_string = value
 		self._parse_idn_string(self._idn_string)
 
+	@property
+	def instr_options(self) -> InstrumentOptions:
+		"""Public getter for the lazy property instr_options"""
+		if self._instr_options is None:
+			self._query_options_and_parse(self.instr_options_parse_mode)
+		return self._instr_options
+
 	@property
 	def opc_timeout(self) -> int:
 		"""See the opc_timeout.setter."""
@@ -129,6 +159,14 @@ class Instrument(object):
 		"""Sets the maximum size of one block transferred during write/read operations."""
 		self._session.data_chunk_size = int(chunk_size)
 
+	# noinspection PyMethodMayBeStatic
+	def _sim_cached_value_found(self, value) -> bool:
+		return value != 'Simulating'
+
+	# noinspection PyMethodMayBeStatic
+	def _sim_cached_value_not_found(self, value) -> bool:
+		return value == 'Simulating'
+
 	def _parse_idn_string(self, idn_string: str) -> None:
 		"""Parse the *IDN? response to:
 		- Manufacturer
@@ -136,7 +174,7 @@ class Instrument(object):
 		- SerialNumber
 		- FirmwareRevision"""
 		idn_string = idn_string.strip()
-		m = re.search(r'([\w& ]+),([a-zA-Z ]+)([\-0-9a-zA-Z ]*),([^,]+),([\w .\-\/]+)', idn_string)
+		m = re.search(r'([\w& ]+),([a-zA-Z ]+)([\-0-9a-zA-Z ]*),([^,]+),([\w .\-/]+)', idn_string)
 		if not m:
 			raise Exception(f"The instrument *IDN? string parsing failed. Parsed *IDN? response: '{idn_string}'")
 		self.manufacturer = m.group(1).strip()
@@ -169,11 +207,13 @@ class Instrument(object):
 	def _query_options_and_parse(self, mode: InstrumentOptions.ParseMode) -> None:
 		"""Queries *OPT? and parses it based on the ParseMode."""
 		if mode == InstrumentOptions.ParseMode.Skip:
-			self.instr_options = InstrumentOptions.Options('', mode)
+			self._instr_options = InstrumentOptions.Options('', mode)
 			return
 		if self._simulating is False:
-			opts = self._session.query_str('*OPT?')
-			self.instr_options = InstrumentOptions.Options(opts, mode)
+			opts = self._session.query_str_no_tout_err('*OPT?', 1000)
+			if opts is None:
+				opts = 'Cannot read the instrument options - *OPT? query is not supported'
+			self._instr_options = InstrumentOptions.Options(opts, mode)
 
 	@staticmethod
 	def _parse_err_query_response(response: str) -> str:
@@ -185,33 +225,31 @@ class Instrument(object):
 		else:
 			return response
 
-	def add_global_repcap(self, name: str, dflt_val: int) -> None:
+	def add_global_repcap(self, name: str, rep_cap: RepeatedCapability) -> None:
 		"""Adds the global repcap name to the list of global repcaps
 		and sets its value to the provided default value."""
 		if name in self._global_repcaps:
 			raise Exception(f"Error adding new global repcap: '{name}' already exists in the list.")
-		self._global_repcaps[name] = dflt_val
+		self._global_repcaps[name] = rep_cap
 
-	def set_global_repcap(self, name: str, value: int) -> None:
-		"""Updates an existing global repcap value."""
+	def set_global_repcap_value(self, name: str, enum_value: Enum) -> None:
+		"""Updates the existing global repcap value as enum"""
 		if name not in self._global_repcaps:
 			raise Exception(f"Error updating global repcap: '{name}' does not exist in the list.")
-		self._global_repcaps[name] = value
+		self._global_repcaps[name].set_enum_value(enum_value)
 
-	def get_global_repcap(self, name: str) -> int:
-		"""Returns the current global repcap name."""
+	def get_global_repcap_value(self, name: str) -> Enum:
+		"""Returns the current global repcap value as enum"""
 		if name not in self._global_repcaps:
 			raise Exception(f"Error retrieving global repcap: '{name}' does not exist in the list.")
-		return self._global_repcaps[name]
+		return self._global_repcaps[name].get_enum_value()
 
 	def _replace_global_repcaps(self, cmd: str) -> str:
 		"""Replaces all the global repcaps in the command: e.g. '<instance>' => '1'.
 		Returns the replaced command."""
 		for name, value in self._global_repcaps.items():
-			if isinstance(value, Enum):
-				value = value.value
-			replace_val = '' if value == 0 else str(value)
-			cmd = cmd.replace(name, replace_val)
+			cmd_value = value.get_cmd_string_value()
+			cmd = cmd.replace(name, cmd_value)
 		return cmd
 
 	def query_opc(self) -> bool:
@@ -234,11 +272,8 @@ class Instrument(object):
 	def check_status(self) -> None:
 		"""Throws InstrumentStatusException in case of an error in the instrument's error queue.
 		The procedure is skipped, if the QueryInstrumentStatus is set to false."""
-		if self._simulating:
-			return
 		if not self.query_instr_status:
 			return
-
 		call_syst_error = self._session.error_in_error_queue() if self.stb_in_error_check else True
 		if call_syst_error:
 			errors = self.query_all_syst_errors(False)
@@ -246,15 +281,11 @@ class Instrument(object):
 
 	def clear_status(self) -> None:
 		"""Clears instrument's status subsystem."""
-		if self._simulating:
-			return
 		self._session.clear()
 		self._session.clear_before_read()
 
 	def reset(self) -> None:
 		"""Resets the instrument and clears its status."""
-		if self._simulating:
-			return
 		self.write("*RST")
 		self.query_opc()
 		self.clear_status()
@@ -262,8 +293,6 @@ class Instrument(object):
 
 	def write(self, cmd: str) -> None:
 		"""Writes string command to the instrument."""
-		if self._simulating:
-			return
 		cmd = self._replace_global_repcaps(cmd)
 		self._session.write(cmd)
 		if self.opc_query_after_write:
@@ -276,8 +305,6 @@ class Instrument(object):
 		"""Writes a OPC-synced command.
 		Also performs error checking if the property self.query_instr_status is set to True.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return
 		cmd = self._replace_global_repcaps(cmd)
 		self._session.write_with_opc(cmd, timeout)
 		self._session.query_and_clear_esr()
@@ -287,8 +314,6 @@ class Instrument(object):
 
 	def write_struct(self, cmd: str, struct: object) -> None:
 		"""Writes command to the instrument with the parameter composed from the entered structure."""
-		if self._simulating:
-			return
 		worker = ArgStructList(struct)
 		param = worker.compose_cmd_string()
 		cmd += f' {param}'.rstrip()
@@ -297,8 +322,6 @@ class Instrument(object):
 	def write_struct_with_opc(self, cmd: str, struct: object, timeout: int = None) -> None:
 		"""Writes OPC-synced command to the instrument with the parameter composed from the entered structure.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return
 		worker = ArgStructList(struct)
 		param = worker.compose_cmd_string()
 		cmd += f' {param}'.rstrip()
@@ -307,10 +330,7 @@ class Instrument(object):
 	def query_str(self, query: str) -> str:
 		"""Sends a query and reads response from the instrument.
 		The response is trimmed of any trailing LF characters and has no length limit."""
-		if self._simulating:
-			return 'Simulating'
 		query = self._replace_global_repcaps(query)
-
 		self.start_send_read_event(query, False)
 		response = self._session.query_str(query)
 		self.end_send_read_event()
@@ -325,14 +345,10 @@ class Instrument(object):
 		Also performs error checking if the self.query_instr_status is true.
 		The response is trimmed of any trailing LF characters and has no length limit.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return 'Simulating'
 		query = self._replace_global_repcaps(query)
-
 		self.start_send_read_event(query, True)
 		response = self._session.query_str_with_opc(query, timeout)
 		self.end_send_read_event()
-
 		if response.endswith('\n'):
 			response = response.strip()
 		self._session.query_and_clear_esr()
@@ -343,8 +359,6 @@ class Instrument(object):
 		"""Queries binary data block to the provided stream.
 		Use it for querying data from instrument to a variable or a file.
 		Throws an exception if the returned data was not a binary data."""
-		if self._simulating:
-			return
 		query = self._replace_global_repcaps(query)
 		self.start_send_read_event(query, False)
 		self._session.query_bin_block(query, stream, True)
@@ -354,8 +368,6 @@ class Instrument(object):
 	def query_bin_block(self, query: str) -> bytes:
 		"""Queries binary data block to bytes and returns data as bytes.
 		Throws an exception if the returned data was not a binary data."""
-		if self._simulating:
-			return bytes([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
 		writer = StreamWriter.as_bin_var()
 		self.query_bin_block_to_stream(query, writer)
 		return writer.content
@@ -372,8 +384,6 @@ class Instrument(object):
 		"""Sends a OPC-synced query and returns data the provided stream.
 		Use it for querying data from instrument to a variable or a file.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return
 		query = self._replace_global_repcaps(query)
 		self.start_send_read_event(query, True)
 		self._session.query_bin_block_with_opc(query, stream, True, timeout)
@@ -384,8 +394,6 @@ class Instrument(object):
 	def query_bin_block_with_opc(self, query: str, timeout: int = None) -> bytes:
 		"""Sends a OPC-synced query and returns data as bytes.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return bytes([0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10])
 		with StreamWriter.as_bin_var() as stream:
 			self.query_bin_block_to_stream_with_opc(query, stream, timeout)
 			return stream.content
@@ -400,73 +408,70 @@ class Instrument(object):
 
 	def query_int(self, query: str) -> int:
 		"""Sends a query and reads response from the instrument as integer."""
-		if self._simulating:
-			return 0
 		string = self.query_str(query)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			return 0
 		return Conv.str_to_int(string)
 
 	def query_int_with_opc(self, query: str, timeout: int = None) -> int:
 		"""Sends a OPC-synced query and reads response from the instrument as integer number.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return 0
 		string = self.query_str_with_opc(query, timeout)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			return 0
 		return Conv.str_to_int(string)
 
 	def query_float(self, query: str) -> float:
 		"""Sends a query and reads response from the instrument as float number."""
-		if self._simulating:
-			return 0.0
 		string = self.query_str(query)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			return 0.0
 		return Conv.str_to_float(string)
 
 	def query_float_with_opc(self, query: str, timeout: int = None) -> float:
 		"""Sends a OPC-synced query and reads response from the instrument as float number.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return 0.0
 		string = self.query_str_with_opc(query, timeout)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			return 0.0
 		return Conv.str_to_float(string)
 
 	def query_bool(self, query: str) -> bool:
 		"""Sends a query and reads response from the instrument as boolean value."""
-		if self._simulating:
-			return False
 		string = self.query_str(query)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			return False
 		return Conv.str_to_bool(string)
 
 	def query_bool_with_opc(self, query: str, timeout: int = None) -> bool:
 		"""Sends a OPC-synced query and reads response from the instrument as boolean value.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return False
 		string = self.query_str_with_opc(query, timeout)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			return False
 		return Conv.str_to_bool(string)
 
 	def query_str_list(self, query: str) -> List[str]:
 		"""Sends a query and reads response from the instrument as csv-list."""
-		if self._simulating:
-			return []
-		response = self.query_str(query).split(',')
-		response = [trim_str_response(x) for x in response]
+		string = self.query_str(query)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			string = 'AAA,BBB,CCC,DDD,EEE,FFF,GGG,HHH,III,JJJ'
+		response = [trim_str_response(x) for x in string.split(',')]
 		return response
 
 	def query_str_list_with_opc(self, query: str, timeout: int = None) -> List[str]:
 		"""Sends a OPC-synced query and reads response from the instrument as csv-list.
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return []
-		response = self.query_str_with_opc(query, timeout).split(',')
-		response = [trim_str_response(x) for x in response]
+		string = self.query_str_with_opc(query, timeout)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			string = 'AAA,BBB,CCC,DDD,EEE,FFF,GGG,HHH,III,JJJ'
+		response = [trim_str_response(x) for x in string.split(',')]
 		return response
 
 	def write_bin_block_from_stream(self, cmd: str, stream: StreamReader) -> None:
 		"""Writes all the stream content as binary data block to the instrument.
 		Use it for writing data to instrument from a variable or a file.
 		The binary data header is added at the beginning of the transmission automatically, do not include it in the payload!!!"""
-		if self._simulating:
-			stream.read()
-			return
 		if self.on_write_handler:
 			self.start_send_write_bin_event(cmd)
 		self._session.write_bin_block(cmd, stream)
@@ -488,13 +493,13 @@ class Instrument(object):
 		"""Queries a list of floating-point numbers that can be returned in ASCII format or in binary format.
 		- For ASCII format, the list numbers are decoded as comma-separated values.
 		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32)."""
-		if self._simulating:
-			return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
 		query = self._replace_global_repcaps(query)
 		self.start_send_read_event(query, False)
 		stream = StreamWriter.as_bin_var()
 		self._session.query_bin_block(query, stream, False)
 		self.end_send_read_event()
+		if self._simulating and not self._session.cached_to_stream:
+			return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
 		if stream.binary:
 			result = Conv.bytes_to_list_of_floats(stream.content, self.bin_float_numbers_format)
 		else:
@@ -507,14 +512,13 @@ class Instrument(object):
 		- For ASCII format, the list numbers are decoded as comma-separated values.
 		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
 		query = self._replace_global_repcaps(query)
-
 		self.start_send_read_event(query, True)
 		stream = StreamWriter.as_bin_var()
 		self._session.query_bin_block_with_opc(query, stream, False, timeout)
 		self.end_send_read_event()
+		if self._simulating and not self._session.cached_to_stream:
+			return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
 		if stream.binary:
 			result = Conv.bytes_to_list_of_floats(stream.content, self.bin_float_numbers_format)
 		else:
@@ -526,34 +530,33 @@ class Instrument(object):
 	def query_bin_or_ascii_float_list_suppressed(self, query: str, suppressed: ArgSingleSuppressed) -> List[float]:
 		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument as list of floats.
 		The current implementation allows for the rest of the string to be only ASCII format."""
-		if self._simulating:
+		string = self.query_str(query)
+		if self._simulating and self._sim_cached_value_not_found(string):
 			return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
-		response = self.query_str(query)
-		response = self._linker.cut_from_response_string(suppressed, response, query)
+		response = self._linker.cut_from_response_string(suppressed, string, query)
 		return Conv.str_to_float_list(response)
 
 	def query_bin_or_ascii_float_list_suppressed_with_opc(self, query: str, suppressed: ArgSingleSuppressed, timeout: int = None) -> List[float]:
 		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument as list of floats.
 		If you do not provide timeout, the method uses current opc_timeout.
 		The current implementation allows for the rest of the string to be only ASCII format."""
-		if self._simulating:
+		string = self.query_str_with_opc(query, timeout)
+		if self._simulating and self._sim_cached_value_not_found(string):
 			return [0.1, 1.2, 2.3, 3.4, 4.5, 5.6, 6.7, 7.8, 8.9, 9.1, 10.2]
-		response = self.query_str_with_opc(query, timeout)
-		response = self._linker.cut_from_response_string(suppressed, response, query)
+		response = self._linker.cut_from_response_string(suppressed, string, query)
 		return Conv.str_to_float_list(response)
 
 	def query_bin_or_ascii_int_list(self, query: str) -> List[int]:
 		"""Queries a list of integer numbers that can be returned in ASCII format or in binary format.
 		- For ASCII format, the list numbers are decoded as comma-separated values.
 		- For Binary Format, the numbers are decoded based on the property BinIntFormat, usually int 32-bit (FORM REAL,32)."""
-		if self._simulating:
-			return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
 		query = self._replace_global_repcaps(query)
-
 		self.start_send_read_event(query, False)
 		stream = StreamWriter.as_bin_var()
 		self._session.query_bin_block(query, stream, False)
 		self.end_send_read_event()
+		if self._simulating and not self._session.cached_to_stream:
+			return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
 		if stream.binary:
 			result = Conv.bytes_to_list_of_integers(stream.content, self.bin_int_numbers_format)
 		else:
@@ -566,14 +569,13 @@ class Instrument(object):
 		- For ASCII format, the list numbers are decoded as comma-separated values.
 		- For Binary Format, the numbers are decoded based on the property BinIntFormat, usually int 32-bit (FORM REAL,32).
 		If you do not provide timeout, the method uses current opc_timeout."""
-		if self._simulating:
-			return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
 		query = self._replace_global_repcaps(query)
-
 		self.start_send_read_event(query, True)
 		stream = StreamWriter.as_bin_var()
 		self._session.query_bin_block_with_opc(query, stream, False, timeout)
 		self.end_send_read_event()
+		if self._simulating and not self._session.cached_to_stream:
+			return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
 		if stream.binary:
 			result = Conv.bytes_to_list_of_integers(stream.content, self.bin_int_numbers_format)
 		else:
@@ -585,9 +587,9 @@ class Instrument(object):
 	def query_bin_or_ascii_int_list_suppressed(self, query: str, suppressed: ArgSingleSuppressed) -> List[int]:
 		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument as list of integers.
 		The current implementation allows for the rest of the string to be only ASCII format."""
+		response = self.query_str(query)
 		if self._simulating:
 			return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
-		response = self.query_str(query)
 		response = self._linker.cut_from_response_string(suppressed, response, query)
 		return Conv.str_to_int_list(response)
 
@@ -595,37 +597,35 @@ class Instrument(object):
 		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument as list of integers.
 		If you do not provide timeout, the method uses current opc_timeout.
 		The current implementation allows for the rest of the string to be only ASCII format."""
+		response = self.query_str_with_opc(query, timeout)
 		if self._simulating:
 			return [1, 2, 3, 5, 10, 15, 20, 30, 50, 100]
-		response = self.query_str_with_opc(query, timeout)
 		response = self._linker.cut_from_response_string(suppressed, response, query)
 		return Conv.str_to_int_list(response)
 
 	def query_struct(self, query: str, struct: object) -> object:
 		"""Queries string of from instrument, and parses it based on the provided structure object.
 		THe method returns the copy of the entered object that it had modified."""
-		if self._simulating:
+		string = self.query_str(query)
+		if self._simulating and self._sim_cached_value_not_found(string):
 			return struct
-		response = self.query_str(query)
 		struct_list = ArgStructList(struct)
-		struct_list.parse_from_cmd_response(response)
+		struct_list.parse_from_cmd_response(string)
 		self._linker.invoke_struct_intern_links(struct, struct_list.args, query)
 		return struct
 
 	def query_str_suppressed(self, query: str, suppressed: ArgSingleSuppressed) -> str:
 		"""Queries string of unknown size from instrument, and returns the part without the suppressed argument."""
-		if self._simulating:
-			return 'simulation'
-		response = self.query_str(query)
-		response = self._linker.cut_from_response_string(suppressed, response, query)
+		string = self.query_str(query)
+		if self._simulating and self._sim_cached_value_not_found(string):
+			return string
+		response = self._linker.cut_from_response_string(suppressed, string, query)
 		return response
 
 	def self_test(self, timeout: int = None) -> Tuple[int, str]:
 		"""Performs instrument's selftest (*TST?).
 		Returns tuple (code:int, message: str). . Code 0 means the self-test passed.
 		You can define the custom timeout in milliseconds. If you do not define it, the default selftest timeout is used (usually 60 secs)."""
-		if self._simulating:
-			return 0, 'Passed'
 		if timeout is None or timeout == 0:
 			timeout = self.selftest_timeout
 		response = self.query_str_with_opc('*TST?', timeout)
@@ -633,7 +633,7 @@ class Instrument(object):
 		if not m:
 			raise InstrumentErrors.UnexpectedResponseException(self.resource_name, f"Unexpected response to a '*TST?' self-test query: '{response}'")
 		code = Conv.str_to_int(m.group(1))
-		msg = m.group(2)
+		msg = Utilities.trim_str_response(m.group(3))
 		self.check_status()
 		return code, msg
 
