@@ -1,14 +1,19 @@
 import time
 from enum import Enum, Flag
 from typing import Tuple, Callable, AnyStr
+import re
 
+# noinspection PyPackageRequirements
 import visa
 
+from .VisaPluginSocketIo import ResourceManager, SocketIo
 from . import InstrumentSettings, InstrumentErrors, Conversions as Conv
 from .InstrumentSettings import WaitForOpcMode, InstrViClearMode as ViClearMode
 from .StreamReader import StreamReader
 from .StreamWriter import StreamWriter
 from .Utilities import size_to_kb_mb_string, calculate_chunks_count
+import platform
+import struct
 
 
 class SessionKind(Enum):
@@ -58,6 +63,11 @@ class VisaSession(object):
 		self._data_chunk_size: int = None
 		self._std_bin_block_header_max_len: int = 999999999
 		self.disable_opc_query: bool = settings.disable_opc_query
+		self.last_status = None
+		self.visa_library_name = None
+
+		# Implemented for interface compatibility with VisaSessionSim
+		self.cached_to_stream = False
 
 		# Event handlers
 		# noinspection PyTypeChecker
@@ -71,16 +81,20 @@ class VisaSession(object):
 
 		if self.reusing_session:
 			# Reuse the session
-			assert isinstance(direct_session, visa.Resource), f"direct_session must be a VISA resource object. Actual type: '{type(visa.Resource)}', value: '{visa.Resource}'"
+			assert isinstance(direct_session, visa.Resource) or isinstance(direct_session, SocketIo), f"Direct_session must be a VISA resource object. Actual type: '{type(direct_session)}', value: '{direct_session}'"
 			self._session = direct_session
-			self.resource_name = resource_name
-			self.resource_name = self._session.resource_name
+			self._resource_name = self._session.resource_name
 		else:
 			# Create new session
-			self._rm = visa.ResourceManager()
-			self.manufacturer = self._rm.visalib.get_attribute(self._rm.session, visa.constants.VI_ATTR_RSRC_MANF_NAME)[0]
+			# Check resource_name for the trailing (SelectVisa=..)
+			pure_resource_name, visa_select = self._get_pure_resource_name(resource_name)
+			if settings.visa_select is not None:
+				visa_select = settings.visa_select
+			self._rm = VisaSession.get_resource_manager(visa_select)
+			self.manufacturer = self._get_visa_manufacturer()
+
 			try:
-				self._session = self._rm.open_resource(resource_name)
+				self._session = self._rm.open_resource(pure_resource_name)
 			except visa.VisaIOError as e:
 				if e.error_code != visa.constants.StatusCode.error_resource_not_found:
 					raise e
@@ -88,8 +102,7 @@ class VisaSession(object):
 				message += f"\nLibrary: {self._rm.visalib}\nManufacturer: {self.manufacturer}\nResource Name: '{resource_name}'"
 				raise InstrumentErrors.ResourceError(resource_name, message)
 
-			self.manufacturer = self._session.resource_manufacturer_name
-			self.resource_name = resource_name
+			self._resource_name = resource_name
 
 		self._interface_type = SessionKind.unsupported
 		if self._session.interface_type == visa.constants.InterfaceType.gpib or self._session.interface_type == visa.constants.InterfaceType.gpib:
@@ -109,7 +122,7 @@ class VisaSession(object):
 
 		elif self._session.interface_type == visa.constants.InterfaceType.tcpip:
 			self._interface_type = SessionKind.vxi11
-			if self._session.resource_class == "SOCKET":
+			if self._session.resource_class == 'SOCKET':
 				self._interface_type = SessionKind.socket
 
 		# Specifics for different interfaces
@@ -128,8 +141,12 @@ class VisaSession(object):
 		else:
 			self._session.read_termination = ''
 
+		# NRP-Z specific settings
 		if self.is_rsnrp_session():
 			self.disable_opc_query = True
+			# NRP-Z does not support chunk reading, therefore the segment must be in one piece
+			settings.io_segment_size = 1000000
+			self.vxi_capable = False
 
 		self.write_delay = settings.write_delay
 		self.read_delay = settings.read_delay
@@ -165,7 +182,49 @@ class VisaSession(object):
 		# Apply settings for ESE and SRE, plus coerce the _opcWaitMode if necessary
 		self._opc_wait_mode = self._set_regs_ese_sre(self._opc_wait_mode)
 
+	@staticmethod
+	def _get_pure_resource_name(resource_name: str):
+		"""Returns pure resource name stripped of the (SelectVisa) part and the visa_select string"""
+		m = re.search(r'(.+)\(SelectVisa=([^),]+)\)', resource_name)
+		if not m:
+			return resource_name, None
+		resource_name = m.group(1).strip()
+		visa_select = m.group(2).strip()
+		return resource_name, visa_select
+
+	@classmethod
+	def get_resource_manager(cls, visa_select: str) -> visa.ResourceManager:
+		"""Returns resource manager for the desired VISA implementation"""
+		operating_system = platform.system().lower()
+		bittness = struct.calcsize("P") * 8
+		if visa_select is None or visa_select in ['@default', '@standard', 'default', 'standard', 'defaultvisa', 'standardvisa', '@defaultvisa', '@standardvisa']:
+			return visa.ResourceManager()
+		if visa_select.lower() in ['@ni', 'ni', 'visa-ni', 'nivisa', 'ni-visa', 'nationalinstruments', 'nationalinstrumentsvisa']:
+			return visa.ResourceManager('@ni')
+		if visa_select.lower() in ['@py', 'pyvisa', 'visa-py', 'pyvisa-py']:
+			return visa.ResourceManager('@py')
+		if 'rohde&schwarz' in visa_select.lower() or 'rohdeschwarz' in visa_select.lower() or visa_select.lower() == 'rsvisa' or visa_select.lower() == 'rs' or visa_select.lower() == 'r&s':
+			if operating_system == 'windows':
+				if bittness == 32:
+					visa_select = r'c:\Windows\SysWOW64\RsVisa32.dll'
+				else:
+					visa_select = r'c:\Windows\system32\RsVisa32.dll'
+				return visa.ResourceManager(visa_select)
+		if visa_select.lower() in ['socketio', 'socket', 'none']:
+			return ResourceManager()
+		return visa.ResourceManager(visa_select)
+
+	def _get_visa_manufacturer(self) -> str:
+		"""Returns manufacturer of the current VISA"""
+		if hasattr(self._rm, 'VisaManufacturerName'):
+			return self._rm.VisaManufacturerName
+		try:
+			return self._rm.visalib.get_attribute(self._rm.session, visa.constants.VI_ATTR_RSRC_MANF_NAME)[0]
+		except TypeError:
+			return self._rm.visalib.__class__.__name__
+
 	def is_rsnrp_session(self) -> bool:
+		"""Returns True, if the current session is a NRP-Z session"""
 		return self._interface_type == SessionKind.rs_nrp
 
 	@property
@@ -273,13 +332,13 @@ class VisaSession(object):
 			if stb & StatusByte.error_queue_not_empty:
 				self.clear()
 				context = f"Sending query '{command.strip()}' with OPC Wait resulted in timeout. OPC Timeout is set to {timeout} ms. Additionally, "
-				InstrumentErrors.assert_no_instrument_status_errors(self.resource_name, self.query_all_syst_errors(), context)
+				InstrumentErrors.assert_no_instrument_status_errors(self._resource_name, self.query_all_syst_errors(), context)
 			InstrumentErrors.throw_opc_tout_exception(self.opc_timeout, timeout, f"Sending query '{command.strip()}'.")
 		else:
 			if stb & StatusByte.error_queue_not_empty:
 				self.clear()
 				context = f"Sending command '{command.strip()}' with OPC Wait resulted in timeout. OPC Timeout is set to {timeout} ms. Additionally, "
-				InstrumentErrors.assert_no_instrument_status_errors(self.resource_name, self.query_all_syst_errors(), context)
+				InstrumentErrors.assert_no_instrument_status_errors(self._resource_name, self.query_all_syst_errors(), context)
 			InstrumentErrors.throw_opc_tout_exception(self.opc_timeout, timeout, f"Sending command '{command.strip()}'.")
 
 	def _narrow_down_io_tout_error(self, context: str) -> None:
@@ -296,7 +355,7 @@ class VisaSession(object):
 				self.visa_timeout = old_tout
 		context = context + f"VISA Timeout error occurred ({self.visa_timeout} milliseconds)"
 		if stb & StatusByte.error_queue_not_empty:
-			InstrumentErrors.assert_no_instrument_status_errors(self.resource_name, self.query_all_syst_errors(), context + " and ...")
+			InstrumentErrors.assert_no_instrument_status_errors(self._resource_name, self.query_all_syst_errors(), context + " and ...")
 		# In case the previous exception is not thrown
 		raise InstrumentErrors.TimeoutException(context)
 
@@ -405,7 +464,7 @@ class VisaSession(object):
 					opc = opc.strip()
 					correct = opc == '0' or opc == '1'
 				if not correct:
-					# read again with a small VISA timeout
+					# Read again with a small VISA timeout
 					opc = self._read_str_timed(5, True)
 				repeat += 1
 				if repeat > 10:
@@ -415,10 +474,12 @@ class VisaSession(object):
 		condition = StatusByte.error_queue_not_empty | StatusByte.message_available | StatusByte.event_status_byte
 		if not stb & condition:
 			return
+		repeat = 0
 		# Loop more times to clear the status subsystem
 		while stb & condition:
 			if stb & StatusByte.error_queue_not_empty:
 				self.write("*CLS")
+				_ = self.query_all_syst_errors()
 			if stb & StatusByte.message_available:
 				# Clear output buffer
 				self._flush_junk_data()
@@ -430,7 +491,9 @@ class VisaSession(object):
 			previous_stb = stb
 			stb = self._query_stb()
 			if stb == previous_stb:
-				raise Exception(f"Cannot clear instrument's status subsystem. Status Byte: '{stb}'")
+				repeat += 1
+				if repeat > 10:
+					raise Exception(f"Cannot clear the instrument's status subsystem. Status Byte: '{stb}'")
 
 	def _flush_junk_data(self) -> None:
 		"""Reads junk bytes to clear the instrument's output buffer."""
@@ -533,8 +596,8 @@ class VisaSession(object):
 
 	def _read_unknown_len(self, stream: StreamWriter, allow_chunk_events: bool, prepend_data: AnyStr = None) -> None:
 		"""Reads data of unknown length to the provided WriteStream.
-		The read is performed in an incremental chunk steps to optimize memory use:
-			- The first read is performed with the fixed size of 1024 bytes.
+		The read is performed in an incremental chunk steps to optimize memory use (for NRP-Z session it is set to fixed self._data_chunk_size):
+			- The first read is performed with the fixed size of 1024 bytes
 			- The 2nd one reads 64 kBytes
 			- The 3rd one reads 128 kBytes
 			- The 4th one reads 256 kBytes and so on, with the max cap of self._data_chunk_size
@@ -546,36 +609,39 @@ class VisaSession(object):
 			if prepend_data and isinstance(prepend_data, str):
 				prepend_data = prepend_data.encode('utf-8')
 			chunk_ix = 0
-			continue_status = visa.constants.StatusCode.success_max_count_read
 			eot = False
 			while not eot:
-				if chunk_ix == 0:
-					# First read, set 1024 bytes read size
-					chunk_size = 1024
-				elif chunk_ix == 1:
-					chunk_size = 65536
+				if self.is_rsnrp_session():
+					chunk_size = self._data_chunk_size
 				else:
-					chunk_size *= 2
+					if chunk_ix == 0:
+						# First read, set 1024 bytes read size
+						chunk_size = 1024
+					elif chunk_ix == 1:
+						chunk_size = 65536
+					else:
+						chunk_size *= 2
 				if chunk_size > self._data_chunk_size:
 					chunk_size = self._data_chunk_size
-
-				chunk, status = self._session.visalib.read(self._session.session, chunk_size)
+				chunk, self.last_status = self._session.visalib.read(self._session.session, chunk_size)
 				if chunk_ix == 0 and prepend_data:
 					chunk = prepend_data + chunk
-				eot = status != continue_status
-
+				eot = not self._last_status_more_data_available()
 				if not stream.binary:
 					chunk = chunk.decode('utf-8')
 					if eot:
 						chunk = chunk.rstrip('\n')
 				stream.write(chunk)
-
 				if self.on_read_chunk_handler and allow_chunk_events:
 					total_size = len(stream) if eot is True else None
 					event_args = EventArgsChunk(stream.binary, chunk_ix, len(chunk), total_size, len(stream), eot, None, chunk if self.io_events_include_data else None)
 					self.on_read_chunk_handler(event_args)
 				chunk_ix += 1
 
+	def _last_status_more_data_available(self):
+		"""Returns True, if the last status signalled that more data is available"""
+		return self.last_status == visa.constants.StatusCode.success_max_count_read
+
 	def _read_str_no_events(self) -> str:
 		"""Reads response from the instrument. The response is then trimmed for trailing LF. \n
 		Sending of any read events is blocked."""
@@ -640,6 +706,20 @@ class VisaSession(object):
 			self._narrow_down_io_tout_error("Querying '{}' - ".format(query.rstrip('\n')))
 		return response
 
+	def query_str_no_tout_err(self, query: str, tout: int) -> str:
+		"""Same as query_str, but you can set the timeout just for this one call.
+		If the timeout exception occurs, it is suppressed and the method returns Null"""
+		response = None
+		old_tout = self.visa_timeout
+		try:
+			self.visa_timeout = tout
+			response = self.query_str(query)
+		except (visa.VisaIOError, InstrumentErrors.StatusException):
+			pass
+		finally:
+			self.visa_timeout = old_tout
+		return response
+
 	def write_with_opc(self, command: str, timeout: int = None) -> None:
 		"""Sends command with OPC-sync.
 		If you do not provide timeout, the method uses current opc_timeout."""
@@ -678,7 +758,7 @@ class VisaSession(object):
 	def query_opc(self) -> bool:
 		"""Sends *OPC? query and reads the result."""
 		if self.disable_opc_query:
-			return 1
+			return True
 		response = self._query_str_no_events("*OPC?")
 		return Conv.str_to_bool(response)
 
@@ -698,7 +778,7 @@ class VisaSession(object):
 		context = context + " SCPI query '{}'".format(query.rstrip('\n'))
 		if stb & StatusByte.error_queue_not_empty:
 			# Instrument reports an error
-			InstrumentErrors.assert_no_instrument_status_errors(self.resource_name, self.query_all_syst_errors(), context)
+			InstrumentErrors.assert_no_instrument_status_errors(self._resource_name, self.query_all_syst_errors(), context)
 		else:
 			# Sometimes even if the StatusByte.MessageAvailable is false, the message is available.
 			# Try to read the STB again
@@ -831,12 +911,12 @@ class VisaSession(object):
 		whole_hdr = char.decode("utf-8")
 		if exc_if_not_bin:
 			if data_type == ReadDataType.null:
-				InstrumentErrors.throw_bin_block_unexp_resp_exception(self.resource_name, '\n')
+				InstrumentErrors.throw_bin_block_unexp_resp_exception(self._resource_name, '\n')
 			# Read 20 more characters to compose a better exception message
 			whole_hdr += self.read_up_to_char(b'\n', 20).decode("utf-8")
-			if self._session.last_status == visa.constants.StatusCode.success_max_count_read:
+			if self.last_status == visa.constants.StatusCode.success_max_count_read:
 				self._flush_junk_data()
-			InstrumentErrors.throw_bin_block_unexp_resp_exception(self.resource_name, whole_hdr)
+			InstrumentErrors.throw_bin_block_unexp_resp_exception(self._resource_name, whole_hdr)
 		return data_type, whole_hdr, length
 
 	def read_bin_block(self, stream: StreamWriter, exc_if_not_bin: bool) -> None:
@@ -850,7 +930,7 @@ class VisaSession(object):
 		elif data_type == ReadDataType.null:
 			# No data, consider it ASCII, Return empty string, and False (signaling ASCII transfer)
 			stream.switch_to_string_data()
-		if data_type == ReadDataType.bin_unknown_len:
+		elif data_type == ReadDataType.bin_unknown_len:
 			if not self.vxi_capable:
 				raise Exception(f'Non-Vxi11 sessions can not read binary data block of unknown length.')
 			self._read_unknown_len(stream, True)
@@ -870,19 +950,20 @@ class VisaSession(object):
 				self._session.read_termination = False
 			# Binary data of known length
 			left_to_read = length
+			self.last_status = visa.constants.StatusCode.success
 			with self._session.ignore_warning(visa.constants.StatusCode.success_max_count_read):
 				chunk_ix = 0
 				total_chunks = calculate_chunks_count(length, self._data_chunk_size)
 				while len(stream) < length:
 					chunk_size = min(self._data_chunk_size, left_to_read)
-					chunk, status = self._session.visalib.read(self._session.session, chunk_size)
+					chunk, self.last_status = self._session.visalib.read(self._session.session, chunk_size)
 					left_to_read -= len(chunk)
 					stream.write(chunk)
 					if self.on_read_chunk_handler:
 						event_args = EventArgsChunk(True, chunk_ix, chunk_size, length, len(stream), left_to_read == 0, total_chunks, chunk if self.io_events_include_data else None)
 						self.on_read_chunk_handler(event_args)
 					chunk_ix += 1
-			if self._session.last_status == visa.constants.StatusCode.success_max_count_read:
+			if self._last_status_more_data_available():
 				if not self.vxi_capable:
 					self._session.read_termination = '\n'
 				self._flush_junk_data()
@@ -897,7 +978,8 @@ class VisaSession(object):
 		:param stream: [StreamWriter] target for the read data. Can be string, bytes, or a file
 		:param exc_if_not_bin: [Boolean] if True, the method throws exception if the received data is not binary"""
 		self.write(query)
-		return self.read_bin_block(stream, exc_if_not_bin)
+		self.read_bin_block(stream, exc_if_not_bin)
+		return
 
 	def query_bin_block_with_opc(self, query: str, stream: StreamWriter, exc_if_not_bin: bool = True, timeout: int = None) -> None:
 		"""Query binary data block with OPC and returns it as byte data.
@@ -934,11 +1016,11 @@ class VisaSession(object):
 		Returns the read data including the stop character."""
 		response = b''
 		for i in range(max_cnt):
-			char = self._session.read_bytes(1, break_on_termchar=True)
+			char, self.last_status = self._session.visalib.read(self._session.session, 1)
 			response += char
 			if char in stop_chars:
 				break
-			if self._session.last_status != visa.constants.StatusCode.success_max_count_read:
+			if self.last_status != visa.constants.StatusCode.success_max_count_read:
 				break
 		return response
 
