@@ -3,18 +3,23 @@ from typing import List, Tuple
 from .Fixed_Files import Events
 from .Internal.Core import Core
 from .Internal.Conversions import BinFloatFormat, BinIntFormat
+from .Internal import Conversions as Conv
+from .Internal.VisaSession import VisaSession
 
 
 class RsInstrument:
-	_driver_version_const = '1.0.0.21'
+	_driver_version_const = '1.4.0.29'
 	_driver_options_const = "SupportedInstrModels = All Rohde & Schwarz Instruments, SupportedIdnPatterns = Rohde(-|&)Schwarz, SimulationIdnString = Rohde&Schwarz*SimulationDevice*100001*" + _driver_version_const
 
 	def __init__(
 			self, resource_name: str, id_query: bool = True, reset: bool = False, options: str = None, direct_session: object = None):
-		"""Initializes new RsCmxNrLteSig session. \n
-		For cleaner code, use the RsCmxNrLteSig.from_existing_session(existing_session_obj)
+		"""Initializes new RsInstrument session. \n
+		For cleaner code, use the RsInstrument.from_existing_session(existing_session_obj)
 		Parameter options tokens examples:
 			- 'Simulate = True' - starts the session in simulation mode. Default: False
+			- 'SelectVisa = 'pyvisa' - uses python implementation of VISA, where you do not need the VISA-C installation
+			- 'SelectVisa = 'rs' - forces usage of Rohde&Schwarz Visa
+			- 'SelectVisa = 'ni' - forces usage of National Instruments Visa
 			- 'QueryInstrumentStatus = False' - same as driver.utilities.instrument_status_checking = False. Default: True.
 			- 'DriverSetup=(WriteDelay = 20, ReadDelay = 5)' - Introduces delay of 20ms before each write and 5ms before each read. Default: 0ms for both
 			- 'DriverSetup=(OpcWaitMode = OpcQuery)' - mode for all the opc-synchronised write/reads. Other modes: StbPolling, StbPollingSlow, StbPollingSuperSlow. Default: StbPolling
@@ -33,7 +38,6 @@ class RsInstrument:
 
 		self._core = Core(resource_name, id_query, reset, RsInstrument._driver_options_const, options, direct_session)
 		self._core.driver_version = RsInstrument._driver_version_const
-		self._add_all_global_repcaps()
 		self._custom_properties_init()
 
 	@classmethod
@@ -51,71 +55,81 @@ class RsInstrument:
 			return f"RsInstrument with session closed"
 
 	def close(self) -> None:
-		"""Closes the active RsInstrument session."""
+		"""Closes the active RsInstrument session"""
 		self._core.io.close()
 
+	@staticmethod
+	def list_resources(expression: str = '?*::INSTR', visa_select: str = None) -> List[str]:
+		"""Finds all the resources defined by the expression
+		:param: expression examples:
+			- '?*' - matches all the available instruments
+			- 'USB::?*' - matches all the USB instruments
+			- "TCPIP::192?*' - matches all the LAN instruments with the IP address starting with 192
+		:param: visa_select - optional parameter selecting a specific VISA. Examples: '@ni', '@rs'
+		"""
+		rm = VisaSession.get_resource_manager(visa_select)
+		resources = rm.list_resources(expression)
+		rm.close()
+		return resources
+
 	def get_session_handle(self):
-		"""Returns the underlying pyvisa session."""
+		"""Returns the underlying pyvisa session"""
 		return self._core.get_session_handle()
 
-	def _add_all_global_repcaps(self) -> None:
-		"""Adds all the repcaps defined as global to the instrument's global repcaps dictionary."""
-		self._core.io.add_global_repcap('<Instance>', 1)
-
 	def _custom_properties_init(self) -> None:
-		"""Adds all the interfaces that are custom for the driver."""
+		"""Adds all the interfaces that are custom for the driver"""
 		self.events = Events.Events(self._core)
 
 	# Utilities part follows - copy it manually fro  the Utilities.py file
 	@property
 	def driver_version(self) -> str:
-		"""Returns the instrument driver version."""
+		"""Returns the instrument driver version"""
 		return self._core.driver_version
 
 	@property
 	def idn_string(self) -> str:
 		"""SCPI Command: *IDN?
-		Returns instrument's Identification string."""
+		Returns instrument's Identification string"""
 		return self._core.io.idn_string
 
 	@property
 	def manufacturer(self) -> str:
-		"""Returns manufacturer of the instrument."""
+		"""Returns manufacturer of the instrument"""
 		return self._core.io.manufacturer
 
 	@property
 	def full_instrument_model_name(self) -> str:
-		"""Returns the current instrument's full name e.g. 'FSW26'."""
+		"""Returns the current instrument's full name e.g. 'FSW26'"""
 		return self._core.io.full_model_name
 
 	@property
 	def instrument_model_name(self) -> str:
-		"""Returns the current instrument's family name e.g. 'FSW'."""
+		"""Returns the current instrument's family name e.g. 'FSW'"""
 		return self._core.io.model
 
 	@property
 	def supported_models(self) -> List[str]:
-		"""Returns a list of the instrument models supported by this instrument driver."""
+		"""Returns a list of the instrument models supported by this instrument driver"""
 		return self._core.supported_instr_models
 
 	@property
 	def instrument_firmware_version(self) -> str:
-		"""Returns instrument's firmware version."""
+		"""Returns instrument's firmware version"""
 		return self._core.io.firmware_version
 
 	@property
 	def instrument_serial_number(self) -> str:
-		"""Returns instrument's serial_number."""
+		"""Returns instrument's serial_number"""
 		return self._core.io.serial_number
 
 	def query_opc(self) -> int:
 		"""SCPI command: *OPC?
-		Queries the instrument's OPC bit and hence it waits until the instrument reports operation complete."""
+		Queries the instrument's OPC bit and hence it waits until the instrument reports operation complete"""
 		return self._core.io.query_opc()
 
 	@property
 	def instrument_status_checking(self) -> bool:
-		"""See the instrument_status_checking.setter."""
+		"""See the instrument_status_checking.setter"""
 		return self._core.io.query_instr_status
 
 	@instrument_status_checking.setter
@@ -129,7 +143,7 @@ class RsInstrument:
 
 	@property
 	def opc_query_after_write(self) -> bool:
-		"""See the opc_query_after_write.setter."""
+		"""See the opc_query_after_write.setter"""
 		return self._core.io.opc_query_after_write
 
 	@opc_query_after_write.setter
@@ -141,26 +155,26 @@ class RsInstrument:
 
 	@property
 	def bin_float_numbers_format(self) -> BinFloatFormat:
-		"""Sets / Gets format of float numbers when transferred as binary data."""
+		"""Sets / Gets format of float numbers when transferred as binary data"""
 		return self._core.io.bin_float_numbers_format
 
 	@bin_float_numbers_format.setter
 	def bin_float_numbers_format(self, value: BinFloatFormat) -> None:
-		"""Sets / Gets format of float numbers when transferred as binary data."""
+		"""Sets / Gets format of float numbers when transferred as binary data"""
 		self._core.io.bin_float_numbers_format = value
 
 	@property
 	def bin_int_numbers_format(self) -> BinIntFormat:
-		"""Sets / Gets format of integer numbers when transferred as binary data."""
+		"""Sets / Gets format of integer numbers when transferred as binary data"""
 		return self._core.io.bin_int_numbers_format
 
 	@bin_int_numbers_format.setter
 	def bin_int_numbers_format(self, value: BinIntFormat) -> None:
-		"""Sets / Gets format of integer numbers when transferred as binary data."""
+		"""Sets / Gets format of integer numbers when transferred as binary data"""
 		self._core.io.bin_int_numbers_format = value
 
 	def clear_status(self) -> None:
-		"""Clears instrument's status system, the session's I/O buffers and the instrument's error queue."""
+		"""Clears instrument's status system, the session's I/O buffers and the instrument's error queue"""
 		return self._core.io.clear_status()
 
 	def query_all_errors(self) -> List[str]:
@@ -171,12 +185,12 @@ class RsInstrument:
 	@property
 	def instrument_options(self) -> List[str]:
 		"""Returns all the instrument options.
-		The options are sorted in the ascending order starting with K-options and continuing with B-options."""
+		The options are sorted in the ascending order starting with K-options and continuing with B-options"""
 		return self._core.io.instr_options.get_all()
 
 	def reset(self) -> None:
 		"""SCPI command: *RST
-		Sends *RST command + calls the clear_status()."""
+		Sends *RST command + calls the clear_status()"""
 		return self._core.io.reset()
 
 	def self_test(self, timeout: int = None) -> Tuple[int, str]:
@@ -188,63 +202,96 @@ class RsInstrument:
 
 	@property
 	def opc_timeout(self) -> int:
-		"""See the opc_timeout.setter."""
+		"""See the opc_timeout.setter"""
 		return self._core.io.opc_timeout
 
 	@opc_timeout.setter
 	def opc_timeout(self, value: int) -> None:
-		"""Sets / Gets timeout in milliseconds for all the operations that use OPC synchronization."""
+		"""Sets / Gets timeout in milliseconds for all the operations that use OPC synchronization"""
 		self._core.io.opc_timeout = value
 
 	@property
 	def visa_timeout(self) -> int:
-		"""See the visa_timeout.setter."""
+		"""See the visa_timeout.setter"""
 		return self._core.io.visa_timeout
 
 	@visa_timeout.setter
 	def visa_timeout(self, value) -> None:
-		"""Sets / Gets visa IO timeout in milliseconds."""
+		"""Sets / Gets visa IO timeout in milliseconds"""
 		self._core.io.visa_timeout = value
 
 	@property
 	def data_chunk_size(self) -> int:
-		"""Returns max chunk size of one data block."""
+		"""Returns max chunk size of one data block"""
 		return self._core.io.data_chunk_size
 
 	@data_chunk_size.setter
 	def data_chunk_size(self, chunk_size: int) -> None:
-		"""Sets the maximum size of one block transferred during write/read operations."""
+		"""Sets the maximum size of one block transferred during write/read operations"""
 		self._core.io.data_chunk_size = chunk_size
 
 	@property
 	def visa_manufacturer(self) -> int:
-		"""Returns the manufacturer of the current VISA session."""
+		"""Returns the manufacturer of the current VISA session"""
 		return self._core.io.visa_manufacturer
 
 	def process_all_commands(self) -> None:
 		"""SCPI command: *WAI
-		Stops further commands processing until all commands sent before *WAI have been executed."""
+		Stops further commands processing until all commands sent before *WAI have been executed"""
 		return self._core.io.write('*WAI')
 
 	def write_str(self, cmd: str) -> None:
-		"""Writes the command to the instrument."""
+		"""Writes the command to the instrument"""
 		self._core.io.write(cmd)
 
+	def write_int(self, cmd: str, param: int) -> None:
+		"""Writes the command to the instrument followed by the integer parameter:
+		e.g.: cmd = 'SELECT:INPUT' param = '2', result command = 'SELECT:INPUT 2'"""
+		self._core.io.write(f'{cmd} {param}')
+
+	def write_int_with_opc(self, cmd: str, param: int, timeout: int = None) -> None:
+		"""Writes the command with OPC to the instrument followed by the integer parameter:
+		e.g.: cmd = 'SELECT:INPUT' param = '2', result command = 'SELECT:INPUT 2'
+		If you do not provide timeout, the method uses current opc_timeout."""
+		self._core.io.write_with_opc(f'{cmd} {param}', timeout)
+
+	def write_float(self, cmd: str, param: float) -> None:
+		"""Writes the command to the instrument followed by the boolean parameter:
+		e.g.: cmd = 'CENTER:FREQ' param = '10E6', result command = 'CENTER:FREQ 10E6'"""
+		self._core.io.write(f'{cmd} {Conv.float_to_str(param)}')
+
+	def write_float_with_opc(self, cmd: str, param: float, timeout: int = None) -> None:
+		"""Writes the command with OPC to the instrument followed by the boolean parameter:
+		e.g.: cmd = 'CENTER:FREQ' param = '10E6', result command = 'CENTER:FREQ 10E6'
+		If you do not provide timeout, the method uses current opc_timeout."""
+		self._core.io.write_with_opc(f'{cmd} {Conv.float_to_str(param)}', timeout)
+
+	def write_bool(self, cmd: str, param: bool) -> None:
+		"""Writes the command to the instrument followed by the boolean parameter:
+		e.g.: cmd = 'OUTPUT' param = 'True', result command = 'OUTPUT ON'"""
+		self._core.io.write(f'{cmd} {Conv.bool_to_str(param)}')
+
+	def write_bool_with_opc(self, cmd: str, param: bool, timeout: int = None) -> None:
+		"""Writes the command with OPC to the instrument followed by the boolean parameter:
+		e.g.: cmd = 'OUTPUT' param = 'True', result command = 'OUTPUT ON'
+		If you do not provide timeout, the method uses current opc_timeout."""
+		self._core.io.write_with_opc(f'{cmd} {Conv.bool_to_str(param)}', timeout)
+
 	def query_str(self, query: str) -> str:
 		"""Sends the query to the instrument and returns the response as string.
 		The response is trimmed of any trailing LF characters and has no length limit."""
 		return self._core.io.query_str(query)
 
 	def query_bool(self, query: str) -> bool:
-		"""Sends the query to the instrument and returns the response as boolean."""
+		"""Sends the query to the instrument and returns the response as boolean"""
 		return self._core.io.query_bool(query)
 
 	def query_int(self, query: str) -> int:
-		"""Sends the query to the instrument and returns the response as integer."""
+		"""Sends the query to the instrument and returns the response as integer"""
 		return self._core.io.query_int(query)
 
 	def query_float(self, query: str) -> float:
-		"""Sends the query to the instrument and returns the response as float."""
+		"""Sends the query to the instrument and returns the response as float"""
 		return self._core.io.query_float(query)
 
 	def write_str_with_opc(self, cmd: str, timeout: int = None) -> None:
@@ -302,6 +349,19 @@ class RsInstrument:
 		If you do not provide timeout, the method uses current opc_timeout."""
 		return self._core.io.query_bin_or_ascii_float_list_with_opc(query, timeout)
 
+	def query_bin_or_ascii_int_list(self, query: str) -> List[int]:
+		"""Queries a list of floating-point numbers that can be returned in ASCII format or in binary format.
+		- For ASCII format, the list numbers are decoded as comma-separated values.
+		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32)."""
+		return self._core.io.query_bin_or_ascii_int_list(query)
+
+	def query_bin_or_ascii_int_list_with_opc(self, query: str, timeout: int = None) -> List[int]:
+		"""Sends a OPC-synced query and reads an list of floating-point numbers that can be returned in ASCII format or in binary format.
+		- For ASCII format, the list numbers are decoded as comma-separated values.
+		- For Binary Format, the numbers are decoded based on the property BinFloatFormat, usually float 32-bit (FORM REAL,32).
+		If you do not provide timeout, the method uses current opc_timeout."""
+		return self._core.io.query_bin_or_ascii_int_list_with_opc(query, timeout)
+
 	# Write / Read to file
 	def query_bin_block_to_file(self, cmd: str, file_path: str, append: bool = False) -> None:
 		"""Queries binary data block to the provided file.
@@ -311,8 +371,8 @@ class RsInstrument:
 		Example for transferring a file from Instrument -> PC:
 		cmd = f"MMEM:DATA? '{INSTR_FILE_PATH}'".
 		Alternatively, use the dedicated methods for this purpose:
-			- send_file_from_pc_to_instrument ()
-			- read_file_from_instrument_to_pc ()"""
+		- send_file_from_pc_to_instrument ()
+		- read_file_from_instrument_to_pc ()"""
 		self._core.io.query_bin_block_to_file(cmd, file_path, append)
 
 	def query_bin_block_to_file_with_opc(self, cmd: str, file_path: str, append: bool = False, timeout: int = None) -> None:
@@ -327,8 +387,8 @@ class RsInstrument:
 		Example for transferring a file from PC -> Instrument:
 		cmd = f"MMEM:DATA '{INSTR_FILE_PATH}',".
 		Alternatively, use the dedicated methods for this purpose:
-			- send_file_from_pc_to_instrument ()
-			- read_file_from_instrument_to_pc ()"""
+		- send_file_from_pc_to_instrument ()
+		- read_file_from_instrument_to_pc ()"""
 		self._core.io.write_bin_block_from_file(cmd, file_path)
 
 	def send_file_from_pc_to_instrument(self, source_pc_file: str, target_instr_file: str) -> None:
@@ -343,3 +403,7 @@ class RsInstrument:
 		Set the append_to_pc_file to True if you want to append the read content to the end of the existing PC file"""
 		cmd = f"MMEM:DATA? '{source_instr_file}'"
 		self.query_bin_block_to_file(cmd, target_pc_file, append_to_pc_file)
+
+	def get_last_sent_cmd(self) -> str:
+		"""Returns the last commands sent to the instrument. Only works in simulation mode"""
+		return self._core.get_last_sent_cmd()
